
.. DO NOT EDIT.
.. THIS FILE WAS AUTOMATICALLY GENERATED BY SPHINX-GALLERY.
.. TO MAKE CHANGES, EDIT THE SOURCE PYTHON FILE:
.. "auto_examples/plot_op_einsum.py"
.. LINE NUMBERS ARE GIVEN BELOW.

.. only:: html

    .. note::
        :class: sphx-glr-download-link-note

        :ref:`Go to the end <sphx_glr_download_auto_examples_plot_op_einsum.py>`
        to download the full example code.

.. rst-class:: sphx-glr-example-title

.. _sphx_glr_auto_examples_plot_op_einsum.py:


.. _l-plot-op-einsum:

Compares implementations of Einsum
==================================

This example compares different equations for function :func:`numpy.einsum`.
It compares *numpy* implementation to a custom implementation,
:epkg:`onnxruntime` implementation and :epkg:`opt-einsum` optimisation.
If available, :epkg:`tensorflow` and :epkg:`pytorch` are included as well.
The custom implementation does not do any transpose.
It uses parallelisation and SIMD optimization when the summation
happens on the last axis of both matrices. It only implements
matrix multiplication. We also measure the improvment made with
function :func:`einsum <onnx_extended.tools.einsum.einsum_fct.einsum>`.

Available optimisation
++++++++++++++++++++++

The code shows which optimisation is used for the custom
implementation, *AVX* or *SSE* and the number of available processors,
equal to the default number of used threads to parallelize.

.. GENERATED FROM PYTHON SOURCE LINES 24-48

.. code-block:: Python


    import logging
    import numpy
    import pandas
    import matplotlib.pyplot as plt
    from onnx import TensorProto
    from onnx.helper import (
        make_model,
        make_graph,
        make_node,
        make_tensor_value_info,
        make_opsetid,
    )
    from onnxruntime import InferenceSession
    from onnx_extended.ext_test_case import measure_time, unit_test_going
    from tqdm import tqdm
    from opt_einsum import contract
    from onnx_extended.tools.einsum.einsum_fct import _einsum

    logging.getLogger("matplotlib.font_manager").setLevel(logging.ERROR)
    logging.getLogger("matplotlib.ticker").setLevel(logging.ERROR)
    logging.getLogger("PIL.PngImagePlugin").setLevel(logging.ERROR)
    logging.getLogger("onnx-extended").setLevel(logging.ERROR)








.. GENERATED FROM PYTHON SOURCE LINES 49-51

Einsum: common code
+++++++++++++++++++

.. GENERATED FROM PYTHON SOURCE LINES 51-214

.. code-block:: Python


    try:
        from tensorflow import einsum as tf_einsum, convert_to_tensor
    except ImportError:
        tf_einsum = None
    try:
        from torch import einsum as torch_einsum, from_numpy
    except ImportError:
        torch_einsum = None


    def build_ort_einsum(equation, op_version=18):  # opset=13, 14, ...
        onx = make_model(
            make_graph(
                [make_node("Einsum", ["x", "y"], ["z"], equation=equation)],
                equation,
                [
                    make_tensor_value_info("x", TensorProto.FLOAT, None),
                    make_tensor_value_info("y", TensorProto.FLOAT, None),
                ],
                [make_tensor_value_info("z", TensorProto.FLOAT, None)],
            ),
            opset_imports=[make_opsetid("", op_version)],
            ir_version=9,
        )
        sess = InferenceSession(onx.SerializeToString(), providers=["CPUExecutionProvider"])
        return lambda x, y: sess.run(None, {"x": x, "y": y})


    def build_ort_decomposed(equation, op_version=18):  # opset=13, 14, ...
        cache = _einsum(
            equation,
            numpy.float32,
            opset=op_version,
            optimize=True,
            verbose=True,
            runtime="python",
        )
        if not hasattr(cache, "onnx_"):
            cache.build()
        sess = InferenceSession(
            cache.onnx_.SerializeToString(), providers=["CPUExecutionProvider"]
        )
        return lambda x, y: sess.run(None, {"X0": x, "X1": y})


    def loop_einsum_eq(fct, equation, xs, ys):
        for x, y in zip(xs, ys):
            fct(equation, x, y)


    def loop_einsum_eq_th(fct, equation, xs, ys):
        for x, y in zip(xs, ys):
            fct(equation, x, y, nthread=-1)


    def loop_einsum(fct, xs, ys):
        for x, y in zip(xs, ys):
            fct(x, y)


    def timeit(stmt, ctx, dim, name):
        obs = measure_time(stmt, div_by_number=True, context=ctx, repeat=5, number=1)
        obs["dim"] = dim
        obs["fct"] = name
        return obs


    def benchmark_equation(equation):
        # equations
        ort_einsum = build_ort_einsum(equation)
        ort_einsum_decomposed = build_ort_decomposed(equation)
        res = []
        for dim in tqdm([8, 16, 32, 64, 100, 128, 200, 256]):  # , 500, 512]):
            if unit_test_going() and dim > 64:
                break
            xs = [numpy.random.rand(2, dim, 12, 64).astype(numpy.float32) for _ in range(5)]
            ys = [numpy.random.rand(2, dim, 12, 64).astype(numpy.float32) for _ in range(5)]

            # numpy
            ctx = dict(
                equation=equation,
                xs=xs,
                ys=ys,
                einsum=numpy.einsum,
                loop_einsum=loop_einsum,
                loop_einsum_eq=loop_einsum_eq,
                loop_einsum_eq_th=loop_einsum_eq_th,
            )
            obs = timeit(
                "loop_einsum_eq(einsum, equation, xs, ys)", ctx, dim, "numpy.einsum"
            )
            res.append(obs)

            # opt-einsum
            ctx["einsum"] = contract
            obs = timeit("loop_einsum_eq(einsum, equation, xs, ys)", ctx, dim, "opt-einsum")
            res.append(obs)

            # onnxruntime
            ctx["einsum"] = ort_einsum
            obs = timeit("loop_einsum(einsum, xs, ys)", ctx, dim, "ort-einsum")
            res.append(obs)

            # onnxruntime decomposed
            ctx["einsum"] = ort_einsum_decomposed
            obs = timeit("loop_einsum(einsum, xs, ys)", ctx, dim, "ort-dec")
            res.append(obs)

            if tf_einsum is not None:
                # tensorflow
                ctx["einsum"] = tf_einsum
                ctx["xs"] = [convert_to_tensor(x) for x in xs]
                ctx["ys"] = [convert_to_tensor(y) for y in ys]
                obs = timeit(
                    "loop_einsum_eq(einsum, equation, xs, ys)", ctx, dim, "tf-einsum"
                )
                res.append(obs)

            if torch_einsum is not None:
                # torch
                ctx["einsum"] = torch_einsum
                ctx["xs"] = [from_numpy(x) for x in xs]
                ctx["ys"] = [from_numpy(y) for y in ys]
                obs = timeit(
                    "loop_einsum_eq(einsum, equation, xs, ys)", ctx, dim, "torch-einsum"
                )
                res.append(obs)

        # Dataframes
        df = pandas.DataFrame(res)
        piv = df.pivot(index="dim", columns="fct", values="average")

        rs = piv.copy()
        for c in ["ort-einsum", "ort-dec", "tf-einsum", "torch-einsum", "opt-einsum"]:
            if c not in rs.columns:
                continue
            rs[c] = rs["numpy.einsum"] / rs[c]
        rs["numpy.einsum"] = 1.0

        # Graphs.
        fig, ax = plt.subplots(1, 2, figsize=(14, 5))
        piv.plot(
            logx=True,
            logy=True,
            ax=ax[0],
            title=f"Einsum benchmark\n{equation} -- (2, N, 12, 64) lower better",
        )
        ax[0].legend(prop={"size": 9})
        rs.plot(
            logx=True,
            logy=True,
            ax=ax[1],
            title="Einsum Speedup, baseline=numpy\n%s -- (2, N, 12, 64)"
            " higher better" % equation,
        )
        ax[1].plot([min(rs.index), max(rs.index)], [0.5, 0.5], "g--")
        ax[1].plot([min(rs.index), max(rs.index)], [2.0, 2.0], "g--")
        ax[1].legend(prop={"size": 9})

        return df, rs, ax









.. GENERATED FROM PYTHON SOURCE LINES 215-228

First equation: bsnh,btnh->bnts
+++++++++++++++++++++++++++++++

The decomposition of this equation without einsum function gives
the following.

 .. gdot::
      :script:

      from onnx_extended.tools.einsum import decompose_einsum_equation
      dec = decompose_einsum_equation(
          'bsnh,btnh->bnts', strategy='numpy', clean=True)
      print(dec.to_dot())

.. GENERATED FROM PYTHON SOURCE LINES 228-235

.. code-block:: Python


    dfs = []
    equation = "bsnh,btnh->bnts"
    df, piv, ax = benchmark_equation(equation)
    df.pivot(index="fct", columns="dim", values="average")
    dfs.append(df)




.. image-sg:: /auto_examples/images/sphx_glr_plot_op_einsum_001.png
   :alt: Einsum benchmark bsnh,btnh->bnts -- (2, N, 12, 64) lower better, Einsum Speedup, baseline=numpy bsnh,btnh->bnts -- (2, N, 12, 64) higher better
   :srcset: /auto_examples/images/sphx_glr_plot_op_einsum_001.png
   :class: sphx-glr-single-img


.. rst-class:: sphx-glr-script-out

 .. code-block:: none

      0%|          | 0/121 [00:00<?, ?it/s]    0.021 rtbest='bsnh,btnh->bnts':   0%|          | 0/121 [00:00<?, ?it/s]    0.021 rtbest='bsnh,btnh->bnts':   1%|          | 1/121 [00:00<00:13,  9.18it/s]    0.012 rtbest='btnh,bsnh->bnst':   1%|          | 1/121 [00:00<00:13,  9.18it/s]    0.0078 rtbest='bnth,bsth->btsn':   1%|          | 1/121 [00:00<00:13,  9.18it/s]    0.0074 rtbest='btsh,bnsh->bsnt':   1%|          | 1/121 [00:00<00:13,  9.18it/s]    0.0072 rtbest='bsth,bnth->btns':   1%|          | 1/121 [00:00<00:13,  9.18it/s]    0.0072 rtbest='bsth,bnth->btns':   6%|▌         | 7/121 [00:00<00:03, 36.81it/s]    0.007 rtbest='bshn,bthn->bhts':   6%|▌         | 7/121 [00:00<00:03, 36.81it/s]     0.007 rtbest='bshn,bthn->bhts':  13%|█▎        | 16/121 [00:00<00:01, 58.13it/s]    0.0069 rtbest='bhst,bnst->bsnh':  13%|█▎        | 16/121 [00:00<00:01, 58.13it/s]    0.0069 rtbest='bhst,bnst->bsnh':  21%|██▏       | 26/121 [00:00<00:01, 71.38it/s]    0.0069 rtbest='bhst,bnst->bsnh':  31%|███       | 37/121 [00:00<00:01, 82.18it/s]    0.0069 rtbest='bhst,bnst->bsnh':  40%|███▉      | 48/121 [00:00<00:00, 87.68it/s]    0.0069 rtbest='bhst,bnst->bsnh':  49%|████▉     | 59/121 [00:00<00:00, 92.26it/s]    0.0069 rtbest='bhst,bnst->bsnh':  58%|█████▊    | 70/121 [00:00<00:00, 96.16it/s]    0.0069 rtbest='bhst,bnst->bsnh':  66%|██████▌   | 80/121 [00:01<00:00, 83.85it/s]    0.0069 rtbest='bhst,bnst->bsnh':  74%|███████▎  | 89/121 [00:01<00:00, 85.08it/s]    0.0069 rtbest='bhst,bnst->bsnh':  82%|████████▏ | 99/121 [00:01<00:00, 88.96it/s]    0.0069 rtbest='bhst,bnst->bsnh':  90%|█████████ | 109/121 [00:01<00:00, 89.25it/s]    0.0069 rtbest='bhst,bnst->bsnh':  98%|█████████▊| 119/121 [00:01<00:00, 90.30it/s]    0.0069 rtbest='bhst,bnst->bsnh': 100%|██████████| 121/121 [00:01<00:00, 82.79it/s]
      0%|          | 0/8 [00:00<?, ?it/s]     12%|█▎        | 1/8 [00:00<00:01,  4.14it/s]     25%|██▌       | 2/8 [00:00<00:01,  5.10it/s]     50%|█████     | 4/8 [00:00<00:00,  4.01it/s]     62%|██████▎   | 5/8 [00:01<00:01,  2.45it/s]     75%|███████▌  | 6/8 [00:02<00:01,  1.67it/s]     88%|████████▊ | 7/8 [00:04<00:00,  1.03it/s]    100%|██████████| 8/8 [00:07<00:00,  1.44s/it]    100%|██████████| 8/8 [00:07<00:00,  1.13it/s]




.. GENERATED FROM PYTHON SOURCE LINES 236-252

Second equation: bshn,bthn->bnts
++++++++++++++++++++++++++++++++

The summation does not happen on the last axis but
on the previous one.
Is it worth transposing before doing the summation...
The decomposition of this equation without einsum function gives
the following.

 .. gdot::
      :script:

      from onnx_extended.tools.einsum import decompose_einsum_equation
      dec = decompose_einsum_equation(
          'bshn,bthn->bnts', strategy='numpy', clean=True)
      print(dec.to_dot())

.. GENERATED FROM PYTHON SOURCE LINES 252-258

.. code-block:: Python


    equation = "bshn,bthn->bnts"
    df, piv, ax = benchmark_equation(equation)
    df.pivot(index="fct", columns="dim", values="average")
    dfs.append(df)




.. image-sg:: /auto_examples/images/sphx_glr_plot_op_einsum_002.png
   :alt: Einsum benchmark bshn,bthn->bnts -- (2, N, 12, 64) lower better, Einsum Speedup, baseline=numpy bshn,bthn->bnts -- (2, N, 12, 64) higher better
   :srcset: /auto_examples/images/sphx_glr_plot_op_einsum_002.png
   :class: sphx-glr-single-img


.. rst-class:: sphx-glr-script-out

 .. code-block:: none

      0%|          | 0/121 [00:00<?, ?it/s]    0.061 rtbest='bshn,bthn->bnts':   0%|          | 0/121 [00:00<?, ?it/s]    0.024 rtbest='bshn,bthn->bnts':   0%|          | 0/121 [00:00<?, ?it/s]    0.024 rtbest='bshn,bthn->bnts':   2%|▏         | 2/121 [00:00<00:05, 19.87it/s]    0.015 rtbest='bthn,bshn->bnst':   2%|▏         | 2/121 [00:00<00:05, 19.87it/s]    0.012 rtbest='bnhs,bths->bstn':   2%|▏         | 2/121 [00:00<00:05, 19.87it/s]    0.0095 rtbest='bnht,bsht->btsn':   2%|▏         | 2/121 [00:00<00:05, 19.87it/s]    0.0092 rtbest='bths,bnhs->bsnt':   2%|▏         | 2/121 [00:00<00:05, 19.87it/s]    0.0087 rtbest='bsht,bnht->btns':   2%|▏         | 2/121 [00:00<00:05, 19.87it/s]    0.0078 rtbest='bsnh,btnh->bhts':   2%|▏         | 2/121 [00:00<00:05, 19.87it/s]    0.0072 rtbest='btnh,bsnh->bhst':   2%|▏         | 2/121 [00:00<00:05, 19.87it/s]    0.0072 rtbest='btnh,bsnh->bhst':   8%|▊         | 10/121 [00:00<00:02, 52.64it/s]    0.0071 rtbest='bsth,bnth->bhns':   8%|▊         | 10/121 [00:00<00:02, 52.64it/s]    0.0071 rtbest='bhnt,bsnt->btsh':   8%|▊         | 10/121 [00:00<00:02, 52.64it/s]    0.0071 rtbest='bhnt,bsnt->btsh':  17%|█▋        | 21/121 [00:00<00:01, 74.63it/s]    0.0071 rtbest='bhnt,bsnt->btsh':  26%|██▌       | 31/121 [00:00<00:01, 82.00it/s]    0.0069 rtbest='nhbs,ntbs->nsth':  26%|██▌       | 31/121 [00:00<00:01, 82.00it/s]    0.0069 rtbest='nhbs,ntbs->nsth':  35%|███▍      | 42/121 [00:00<00:00, 89.43it/s]    0.0069 rtbest='nhbs,ntbs->nsth':  42%|████▏     | 51/121 [00:00<00:00, 89.27it/s]    0.0069 rtbest='nhbs,ntbs->nsth':  51%|█████     | 62/121 [00:00<00:00, 93.57it/s]    0.0069 rtbest='nhbs,ntbs->nsth':  60%|██████    | 73/121 [00:00<00:00, 97.66it/s]    0.0069 rtbest='nhbs,ntbs->nsth':  69%|██████▊   | 83/121 [00:00<00:00, 96.16it/s]    0.0069 rtbest='nhbs,ntbs->nsth':  78%|███████▊  | 94/121 [00:01<00:00, 99.53it/s]    0.0069 rtbest='nhbs,ntbs->nsth':  86%|████████▌ | 104/121 [00:01<00:00, 90.51it/s]    0.0069 rtbest='nhbs,ntbs->nsth':  94%|█████████▍| 114/121 [00:01<00:00, 89.39it/s]    0.0069 rtbest='nhbs,ntbs->nsth': 100%|██████████| 121/121 [00:01<00:00, 88.26it/s]
      0%|          | 0/8 [00:00<?, ?it/s]     12%|█▎        | 1/8 [00:00<00:02,  3.11it/s]     25%|██▌       | 2/8 [00:00<00:01,  4.32it/s]     50%|█████     | 4/8 [00:01<00:01,  3.45it/s]     62%|██████▎   | 5/8 [00:01<00:01,  2.17it/s]     75%|███████▌  | 6/8 [00:03<00:01,  1.42it/s]     88%|████████▊ | 7/8 [00:05<00:01,  1.20s/it]    100%|██████████| 8/8 [00:09<00:00,  2.04s/it]    100%|██████████| 8/8 [00:09<00:00,  1.19s/it]




.. GENERATED FROM PYTHON SOURCE LINES 259-274

Third equation: bhsn,bhtn->bnts
+++++++++++++++++++++++++++++++

The summation does not happen on the last axis but
on the second one. It is worth transposing before multiplying.
The decomposition of this equation without einsum function gives
the following.

 .. gdot::
      :script:

      from onnx_extended.tools.einsum import decompose_einsum_equation
      dec = decompose_einsum_equation(
          'bhsn,bhtn->bnts', strategy='numpy', clean=True)
      print(dec.to_dot())

.. GENERATED FROM PYTHON SOURCE LINES 274-280

.. code-block:: Python


    equation = "bhsn,bhtn->bnts"
    df, piv, ax = benchmark_equation(equation)
    df.pivot(index="fct", columns="dim", values="average")
    dfs.append(df)




.. image-sg:: /auto_examples/images/sphx_glr_plot_op_einsum_003.png
   :alt: Einsum benchmark bhsn,bhtn->bnts -- (2, N, 12, 64) lower better, Einsum Speedup, baseline=numpy bhsn,bhtn->bnts -- (2, N, 12, 64) higher better
   :srcset: /auto_examples/images/sphx_glr_plot_op_einsum_003.png
   :class: sphx-glr-single-img


.. rst-class:: sphx-glr-script-out

 .. code-block:: none

      0%|          | 0/121 [00:00<?, ?it/s]    0.061 rtbest='bhsn,bhtn->bnts':   0%|          | 0/121 [00:00<?, ?it/s]    0.034 rtbest='bhsn,bhtn->bnts':   0%|          | 0/121 [00:00<?, ?it/s]    0.034 rtbest='bhsn,bhtn->bnts':   2%|▏         | 2/121 [00:00<00:06, 19.28it/s]    0.033 rtbest='bhtn,bhsn->bnst':   2%|▏         | 2/121 [00:00<00:06, 19.28it/s]    0.02 rtbest='bhns,bhts->bstn':   2%|▏         | 2/121 [00:00<00:06, 19.28it/s]     0.0093 rtbest='bhnt,bhst->btsn':   2%|▏         | 2/121 [00:00<00:06, 19.28it/s]    0.0093 rtbest='bhnt,bhst->btsn':   7%|▋         | 8/121 [00:00<00:02, 40.33it/s]    0.008 rtbest='bnth,bnsh->bhst':   7%|▋         | 8/121 [00:00<00:02, 40.33it/s]     0.0078 rtbest='bsnh,bsth->bhtn':   7%|▋         | 8/121 [00:00<00:02, 40.33it/s]    0.0073 rtbest='btnh,btsh->bhsn':   7%|▋         | 8/121 [00:00<00:02, 40.33it/s]    0.0073 rtbest='btnh,btsh->bhsn':  13%|█▎        | 16/121 [00:00<00:01, 54.50it/s]    0.0073 rtbest='btnh,btsh->bhsn':  21%|██▏       | 26/121 [00:00<00:01, 68.96it/s]    0.0073 rtbest='btnh,btsh->bhsn':  27%|██▋       | 33/121 [00:00<00:01, 62.29it/s]    0.0073 rtbest='btnh,btsh->bhsn':  34%|███▍      | 41/121 [00:00<00:01, 65.19it/s]    0.0073 rtbest='btnh,btsh->bhsn':  40%|███▉      | 48/121 [00:00<00:01, 63.96it/s]    0.0073 rtbest='btnh,btsh->bhsn':  47%|████▋     | 57/121 [00:00<00:00, 69.97it/s]    0.0073 rtbest='btnh,btsh->bhsn':  54%|█████▎    | 65/121 [00:01<00:00, 72.26it/s]    0.0073 rtbest='btnh,btsh->bhsn':  60%|██████    | 73/121 [00:01<00:00, 67.84it/s]    0.0073 rtbest='btnh,btsh->bhsn':  66%|██████▌   | 80/121 [00:01<00:00, 67.53it/s]    0.0073 rtbest='btnh,btsh->bhsn':  72%|███████▏  | 87/121 [00:01<00:00, 61.57it/s]    0.0073 rtbest='btnh,btsh->bhsn':  79%|███████▊  | 95/121 [00:01<00:00, 66.26it/s]    0.0073 rtbest='btnh,btsh->bhsn':  85%|████████▌ | 103/121 [00:01<00:00, 68.67it/s]    0.0073 rtbest='btnh,btsh->bhsn':  92%|█████████▏| 111/121 [00:01<00:00, 67.07it/s]    0.0073 rtbest='btnh,btsh->bhsn':  98%|█████████▊| 118/121 [00:01<00:00, 67.78it/s]    0.0073 rtbest='btnh,btsh->bhsn': 100%|██████████| 121/121 [00:01<00:00, 65.03it/s]
      0%|          | 0/8 [00:00<?, ?it/s]     12%|█▎        | 1/8 [00:00<00:02,  2.88it/s]     25%|██▌       | 2/8 [00:00<00:01,  4.64it/s]     38%|███▊      | 3/8 [00:00<00:01,  3.64it/s]     50%|█████     | 4/8 [00:01<00:01,  3.15it/s]     62%|██████▎   | 5/8 [00:01<00:01,  2.58it/s]     75%|███████▌  | 6/8 [00:02<00:00,  2.07it/s]     88%|████████▊ | 7/8 [00:03<00:00,  1.58it/s]    100%|██████████| 8/8 [00:04<00:00,  1.26it/s]    100%|██████████| 8/8 [00:04<00:00,  1.80it/s]




.. GENERATED FROM PYTHON SOURCE LINES 281-287

Conclusion
++++++++++

pytorch seems quite efficient on these examples.
The custom implementation was a way to investigate
the implementation of einsum and find some ways to optimize it.

.. GENERATED FROM PYTHON SOURCE LINES 287-295

.. code-block:: Python


    merged = pandas.concat(dfs)
    name = "einsum"
    merged.to_csv(f"plot_{name}.csv", index=False)
    merged.to_excel(f"plot_{name}.xlsx", index=False)
    plt.savefig(f"plot_{name}.png")

    # plt.show()



.. image-sg:: /auto_examples/images/sphx_glr_plot_op_einsum_004.png
   :alt: plot op einsum
   :srcset: /auto_examples/images/sphx_glr_plot_op_einsum_004.png
   :class: sphx-glr-single-img






.. rst-class:: sphx-glr-timing

   **Total running time of the script:** (0 minutes 27.314 seconds)


.. _sphx_glr_download_auto_examples_plot_op_einsum.py:

.. only:: html

  .. container:: sphx-glr-footer sphx-glr-footer-example

    .. container:: sphx-glr-download sphx-glr-download-jupyter

      :download:`Download Jupyter notebook: plot_op_einsum.ipynb <plot_op_einsum.ipynb>`

    .. container:: sphx-glr-download sphx-glr-download-python

      :download:`Download Python source code: plot_op_einsum.py <plot_op_einsum.py>`

    .. container:: sphx-glr-download sphx-glr-download-zip

      :download:`Download zipped: plot_op_einsum.zip <plot_op_einsum.zip>`


.. only:: html

 .. rst-class:: sphx-glr-signature

    `Gallery generated by Sphinx-Gallery <https://sphinx-gallery.github.io>`_
